import email, smtplib, ssl
from email import encoders
from email.message import EmailMessage
from email.mime.base import MIMEBase
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
from email.mime.image import MIMEImage
from email.utils import formataddr
import mimetypes
from datetime import datetime
import os , csv , re, socket, requests, string, random

class email_sender :
    def __init__(self, user, password, servername, port = 0, timeout = 30, sender_name = None):
        """Store username , password and servername"""
        self.servername = servername
        self.password = password
        self.sender = user
        self.port, self.timeout = port, timeout
        self.sender_name = sender_name
        
    def executor(self,recipients, message ):    
        """Connect to Server and Send Email, will return status of sending in text""" 
        #Create Connection
        smtpObj = smtplib.SMTP(self.servername, port = self.port, timeout = self.timeout)    
        smtpObj.login(self.sender.split('@')[0], self.password)
        ssl.create_default_context()
        try :
            #Send and Quit
            smtpObj.sendmail(self.sender, recipients, message.as_string())
            smtpObj.quit()
            return 'Send OK'
        except Exception as e:
            #Quit if error
            smtpObj.quit()
            return str(e)

    def add_file(self, message, attachment_path) :
        if os.path.exists(attachment_path):    
            with open(attachment_path, "rb") as attachment:
                # Add file as application/octet-stream
                # Email client can usually download this automatically as attachment
                base_attach = MIMEBase("application", "octet-stream")
                base_attach.set_payload(attachment.read())
            # Encode file in ASCII characters to send by email    
            encoders.encode_base64(base_attach)
            # Add header as key/value pair to attachment part
            base_attach.add_header(
                "Content-Disposition",
                f"attachment; filename= "+ attachment_path.split('/')[-1],)        
            # Add attachment to message and convert message to string
            message.attach(base_attach)
        return message

    def send(self, sendto,subject,text_in , attachment = None, sender_name = None):
        """Create Message Part"""
        #split for multple
        recipients = sendto.split(";")

        #Create Head of Email
        message = MIMEMultipart()
        #message = EmailMessage()
        if sender_name == None : sender_name = self.sender_name
        if sender_name == None : message["From"] = self.sender
        else : message['From'] = formataddr((str(sender_name), self.sender))
        message["To"] =  ', '.join(recipients)
        message["Subject"] = subject

        #Create Message Part
        html = """<html><body><p>{}</p></body></html>""".format(text_in)
        #Attach before send
        partHTML = MIMEText(html, 'html')
        message.attach(partHTML)

        if attachment == None : pass
        elif type(attachment) != list : raise Exception("attachment must be a list")
        else : 
            for i in attachment : 
                message = self.add_file(message , i)
        #Sending Email and return status text
        return self.executor(recipients, message)

class lazy_LINE :
    def __init__(self, token, timeout = 60) :
        self.token = token
        self.timeout = timeout
    
    def help(self) :
        helper = 'Main Page : https://notify-bot.line.me/en\n'
        helper += 'Doc : https://notify-bot.line.me/doc/en\n'
        helper += 'Sticker List : https://developers.line.biz/en/docs/messaging-api/sticker-list'
        print(helper)

    def send(self, message , stickerPackageId = '', stickerId = '', notification = True, picture = '' , timeout = None) :
        if timeout == None : timeout = self.timeout
        payload = {'message' : message , 'notificationDisabled' : not notification}

        if (stickerPackageId != '') & (stickerId != '') :
            payload['stickerPackageId'] = stickerPackageId
            payload['stickerId'] = stickerId
        
        if picture == '' : 
            r = requests.post('https://notify-api.line.me/api/notify'
                                , headers={'Authorization' : 'Bearer {}'.format(self.token)}
                                , params = payload, timeout = timeout)
        else :
            r = requests.post('https://notify-api.line.me/api/notify'
                                , headers={'Authorization' : 'Bearer {}'.format(self.token)}
                                , params = payload , timeout = timeout
                                , files = {'imageFile': open(picture, 'rb')})
        return r

def log_csv(file_name,msg_in):
    logic = True
    if os.path.exists(file_name) : logic = False
    if '/' in file_name : 
        if file_name.split('/')[0] != '' :
            dir_name = re.sub('(/.*\..*)','',file_name)
            if not os.path.isdir(dir_name) : os.mkdir(dir_name)
    with open(file_name, 'a', newline='') as csvfile:
        fieldnames = ['Timestamp','msg']
        writer = csv.DictWriter(csvfile, fieldnames=fieldnames)
        if logic : writer.writeheader()
        writer.writerow({'Timestamp':str(datetime.now()),'msg':msg_in})

def check_port(ip = None, port = None):
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)

    if ip == None : ip = input("Please Input IP :")
    if port == None : port = input("Please Input port :")

    location = (str(ip), int(port))
    r = sock.connect_ex(location)
    if r == 0:
        print("Port is open")
        return True
    else:
        print("Port is not open")
        return False

def check_utc(utc_target) :
    current_utc = (datetime.now() - datetime.utcnow())
    current_utc = current_utc.total_seconds()/3600
    return int(round(utc_target - current_utc,0))

def healthcheck(cpu_interval = 1):
    import shutil, psutil
    pc_name = str(socket.gethostname())
    pc_ip = str(socket.gethostbyname(socket.gethostname()))
    du = shutil.disk_usage("/")
    rom_use = round(100 - du.free/du.total*100,1)
    cpu_use = psutil.cpu_percent(cpu_interval)
    ram_use = psutil.virtual_memory().percent

    return {'hostname' : [pc_name] ,
             'hostip' : [pc_ip],
             'cpu_use': [cpu_use], 
             'ram_use': [ram_use],
             'rom_use': [rom_use]}



def timeout(timeout):
    """
    Refer to https://stackoverflow.com/questions/21827874/timeout-a-function-windows
    
    Usage
    
    func = timeout(timeout=XXX)(My_Function)
    try: func()
    except: pass

    or

    @timeout(XX)
    def My_Function :
        ....
    """
    from threading import Thread
    import functools
    def deco(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            res = [Exception('function [%s] timeout [%s seconds] exceeded!' % (func.__name__, timeout))]
            def newFunc():
                try:
                    res[0] = func(*args, **kwargs)
                except Exception as e:
                    res[0] = e
            t = Thread(target=newFunc)
            t.daemon = True
            try:
                t.start()
                t.join(timeout)
            except Exception as je:
                print ('error starting thread')
                raise je
            ret = res[0]
            if isinstance(ret, BaseException):
                raise ret
            return ret
        return wrapper
    return deco

def random_text(n) :
    """Generate random text 'n' characters"""
    return ''.join([random.choice(string.digits + string.ascii_letters) for x in range(n)])



from opcua import Client

class lazy_OPCUA_connection :
    def __init__(self, client) :
        self.client = client
    
    def __enter__(self) : 
        self.client.connect()
        return self.client

    def __exit__(self, exc_type, exc_val, exc_tb) :
        self.client.disconnect()

class lazy_OPCUA :
    def __init__(self, url, user = '', password = '') :
        self.client = Client(url)
        if user != '' : self.client.set_user(user)
        if password != '' : self.client.set_password(password)
        self.con = lazy_OPCUA_connection(self.client)
   
    def read(self, node, full_value = False) :
        if type(node) == list : multiple_node, out = True, []
        else : multiple_node = False
        with self.con as con :
            if multiple_node :
                for i in node :
                    var = con.get_node(i)
                    if full_value : out.append(var.get_data_value())
                    else : out.append(var.get_value())
            else : 
                var = con.get_node(node)
                if full_value : out = var.get_data_value()
                else : out = var.get_value()
        return out
    
    def write(self, node, value) :
        if (type(node) == list) & (type(value) == list) : 
            assert len(node) == len(value), "Number of nodes and values must be to same!!!"
            multiple_node, out = True, []
        else : multiple_node = False
        with self.con as con :
            if multiple_node :
                i, size = 0, len(node)
                while i < size :
                    var = con.get_node(node[i])
                    var.set_value(value[i])
                    out.append(value[i])
                    i += 1
            else : 
                var = con.get_node(node)
                var.set_value(value)
        return value



import os
from google.cloud import storage
from glob import glob
from datetime import timedelta

class lazy_GCS :
    def __init__(self, project_id, bucket_name,credential = '') :
        self.project_id = project_id
        self.bucket_name = bucket_name
        self.credentials = credential

    def list_folder(self, bucket_folder ,as_blob = False, include_self= False
                    , get_file = True, get_folder = False, all_file = False ) :
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials
        # Initialise a client
        client = storage.Client(self.project_id)
        # Create a bucket object for our bucket
        bucket = client.get_bucket(self.bucket_name)
        # list all objects in the directory
        blobs = bucket.list_blobs(prefix=bucket_folder)
        files, folders = [], []
        for blob in blobs :
            if (blob.name.endswith("/")) : 
                if (blob.name == f'{bucket_folder}/') : source = [blob]
                else : folders.append(blob) # Folder
            else : files.append(blob) # File
        if not all_file :
            for folder in folders : files = [file for file in files if not file.name.startswith(folder.name)]
        out_put = [file for file in files if get_file] + [folder for folder in folders if get_folder]
        if include_self : out_put += source
        if as_blob : return [blob for blob in out_put]
        else : return [blob.name for blob in out_put]
        
    def generate_signed_url(self, bucket_name, object_name, blob = None, expiration = 60):
        """
        Reference from https://cloud.google.com/storage/docs/access-control/signing-urls-with-helpers#code-samples
        """
        if blob == None :
            storage_client = storage.Client()
            bucket = storage_client.bucket(bucket_name)
            blob = bucket.blob(object_name)
        url = blob.generate_signed_url(
                version="v4",
                expiration=timedelta(seconds=expiration),
                method="GET")
        return url

    def download(self, bucket_file, local_file):
        """Need 2 permissions Storage Legacy Object Reader and Storage Legacy Bucket Writer"""
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials
        # Initialise a client
        client = storage.Client(self.project_id)
        # Create a bucket object for our bucket
        bucket = client.get_bucket(self.bucket_name)
        # Create a blob object from the filepath
        blob = bucket.blob(bucket_file)
        # Download the file to a destination
        blob.download_to_filename(local_file)

    def download_folder(self, bucket_folder, local_folder, create_folder = False):
        """Need 2 permissions Storage Legacy Object Reader and Storage Legacy Bucket Writer"""
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials
        if not os.path.isdir(local_folder) :
            if create_folder :
                os.mkdir(local_folder)
            else : print('Local Folder Not Found, Create Folder Manual or add create_folder = True')
        # Initialise a client
        client = storage.Client(self.project_id)
        bucket = client.get_bucket(self.bucket_name)
        blobs = bucket.list_blobs(prefix=bucket_folder)  # Get list of files
        for blob in blobs:
            if blob.name.endswith("/"):
                continue
            filename = blob.name.split('/')[-1]
            blob.download_to_filename(os.path.join(local_folder,filename))  # Download

    def upload(self, bucket_file, local_file , remove_file = False, generate_signed_url = False, url_expiration = 60 ):
        """Need a permission Storage Legacy Bucket Writer"""
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials

        # Initialise a client
        client = storage.Client(self.project_id)
        # Create a bucket object for our bucket
        bucket = client.get_bucket(self.bucket_name)
        # Create a blob object from the filepath
        blob = bucket.blob(bucket_file)
        # Download the file to a destination
        blob.upload_from_filename(local_file)
        if remove_file : os.remove(local_file)
        if generate_signed_url :
            return self.generate_signed_url(  bucket_name = self.bucket_name
                                            , object_name = bucket_file
                                            , blob = blob
                                            , expiration = url_expiration)


    def upload_folder(self, bucket_folder, local_folder , remove_file = False, generate_signed_url = False, url_expiration = 60 ):
        """Need a permission Storage Legacy Bucket Writer"""
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials

        # Initialise a client
        client = storage.Client(self.project_id)
        # Create a bucket object for our bucket
        bucket = client.get_bucket(self.bucket_name)
        # Create List of signed_url if used
        if generate_signed_url : link_list = []
        # Upload File from List
        for i in glob(local_folder + '/*') :
            # Create File Name
            bucket_file = i.replace(local_folder,bucket_folder).replace('\\','/')
            # Create a blob object from the filepath
            blob = bucket.blob(bucket_file)
            # Download the file to a destination
            blob.upload_from_filename(i)
            if remove_file : os.remove(i)
            if generate_signed_url :
                link_list.append(self.generate_signed_url(  bucket_name = self.bucket_name
                                            , object_name = bucket_file
                                            , blob = blob
                                            , expiration = url_expiration))
        if generate_signed_url : return link_list

    def delete(self, bucket_file):
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials
        client = storage.Client(self.project_id)
        bucket = client.get_bucket(self.bucket_name)
        blob = bucket.blob(bucket_file)
        blob.delete()
            
    def delete_folder(self, bucket_folder , delete_folder = False, deep_delete = False):
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials
        # Get blobs
        blobs = self.list_folder(bucket_folder = bucket_folder 
                                , as_blob = True, get_file = True, get_folder = deep_delete
                                , all_file = deep_delete, include_self = delete_folder)
        for blob in blobs : blob.delete()

class da_tran_bucket:
    def __init__(self, project_id, bucket_name,credential = '') :
        self.project_id = project_id
        self.bucket_name = bucket_name
        self.credentials = credential

    def download(self, bucket_file, local_file):
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials

        # Initialise a client
        client = storage.Client(self.project_id)
        # Create a bucket object for our bucket
        bucket = client.get_bucket(self.bucket_name)
        # Create a blob object from the filepath
        blob = bucket.blob(bucket_file)
        # Download the file to a destination
        blob.download_to_filename(local_file)

    def download_folder(self, bucket_folder, local_folder, create_folder = False):
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials
        if not os.path.isdir(local_folder) :
            if create_folder :
                os.mkdir(local_folder)
            else : print('Local Folder Not Found, Create Folder Manual or add create_folder = True')
        # Initialise a client
        client = storage.Client(self.project_id)
        bucket = client.get_bucket(self.bucket_name)
        blobs = bucket.list_blobs(prefix=bucket_folder)  # Get list of files
        for blob in blobs:
            if blob.name.endswith("/"):
                continue
            filename = blob.name.split('/')[-1]
            blob.download_to_filename(os.path.join(local_folder,filename))  # Download

    def upload(self, bucket_file, local_file , remove_file = False):
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials

        # Initialise a client
        client = storage.Client(self.project_id)
        # Create a bucket object for our bucket
        bucket = client.get_bucket(self.bucket_name)
        # Create a blob object from the filepath
        blob = bucket.blob(bucket_file)
        # Download the file to a destination
        blob.upload_from_filename(local_file)
        if remove_file : os.remove(local_file)

    def upload_folder(self, bucket_folder, local_folder , remove_file = False):
        if self.credentials == '' : pass
        else : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials

        # Initialise a client
        client = storage.Client(self.project_id)
        # Create a bucket object for our bucket
        bucket = client.get_bucket(self.bucket_name)
        # Upload File from List
        for i in glob(local_folder + '/*') :
            # Create File Name
            bucket_file = i.replace(local_folder,bucket_folder).replace('\\','/')
            # Create a blob object from the filepath
            blob = bucket.blob(bucket_file)
            # Download the file to a destination
            blob.upload_from_filename(i)
            if remove_file : os.remove(i)



from tensorflow.keras.models import load_model
from tensorflow import device, reduce_mean
from tensorflow.image import ssim
import numpy as np
import cv2

def SSIMLoss(y_true, y_pred):
    return 1 - reduce_mean(ssim(y_true, y_pred,1.0))

class lazy_ImageProcessing :
    """Intented to use with Google TeachableAI's Image Classification and Auto-Encoder"""
    def __init__(self , model_path, model_type = None, class_list = None, loss_function = None
                 , custom_objects = None
                 , size = (224,224,3), cvtColor = None
                 , normalize_type = np.float32
                 , normalize_divide = 127, normalize_add = -1
                 , device = 'cpu') :
        if model_type == 'image' : 
            self.class_list, self.model_type = class_list, 'image'
        elif model_type == 'autoencoder' : 
            self.loss_function, self.model_type = loss_function, 'autoencoder'
        # Load Model
        if custom_objects == None :
            self.model = load_model(model_path)
        else :
            self.model = load_model(model_path, custom_objects= custom_objects)
        self.cvtColor = cvtColor
        self.loss_function, self.size = loss_function, size
        self.normalize_type = normalize_type
        self.normalize_divide, self.normalize_add = normalize_divide, normalize_add
        self.device = device

    def format_image(self, image) :
        image_array, color_ = image[:], 3
        # Recolor
        if self.cvtColor != None : 
            image_array = cv2.cvtColor(image_array, self.cvtColor)
            if (cv2.COLOR_BGR2GRAY == self.cvtColor)|(cv2.COLOR_RGB2GRAY == self.cvtColor) :
                color_ = 1
        # Reshape
        image_array = cv2.resize(image_array, (self.size[0], self.size[1]))
        image_array = np.reshape(image_array , (1,self.size[0], self.size[1], color_))
        # Normalize the image
        normalized_image_array = (image_array.astype(self.normalize_type) / self.normalize_divide) + self.normalize_add
        return normalized_image_array
    
    def predict(self, image, output_type = 'image') :
        with device('/{}:0'.format(self.device)): 
            formatted_image = self.format_image(image)
            prediction = self.model.predict(formatted_image)
        if self.model_type == 'image' : return dict(zip(self.class_list, prediction.flatten().tolist())) 
        elif self.model_type == 'autoencoder' :
            if output_type == 'image' : return cv2.cvtColor(prediction[0], cv2.COLOR_GRAY2RGB)
            elif output_type == 'loss' : 
                return self.loss_function(prediction, formatted_image)
            elif output_type == 'both' : 
                return { 'image' : cv2.cvtColor(prediction[0], cv2.COLOR_GRAY2RGB)
                        ,'loss' : self.loss_function(prediction, formatted_image)}


## FastAPI Core
from fastapi import FastAPI, Body, Request
from fastapi.responses import JSONResponse
import os, traceback, uvicorn
## Security Basic Auth
import secrets, base64
from fastapi import Depends, FastAPI, HTTPException, status
from fastapi.security import HTTPBasic, HTTPBasicCredentials
## Add Doc Security
from fastapi.openapi.docs import get_redoc_html, get_swagger_ui_html
from fastapi.openapi.utils import get_openapi
# Get Query Parameter
# from pydantic import create_model

class lazy_API :
    def __init__(self, title, version, description
                     , username = 'user'
                     , password = 'password'
                     , api_weak_authen = False
                     , callback = None
                ) :
        """lazy_API class, to create FastAPI with a few lines of code"""
        ## Create Fast API
        self.app = FastAPI(   title= title
                            , version= version
                            , description= description
                            , docs_url=None
                            , redoc_url=None
                            , openapi_url = None)     
        ## Initial Basic Auth
        self.storage = [base64.b64encode(username.encode('ascii')),
                        base64.b64encode(password.encode('ascii'))]
        self.api_weak_authen = api_weak_authen
        ## Add Default Callback
        self.callback = callback

    ## Basic Authen
    def basic_authen_check(self, credentials: HTTPBasicCredentials = Depends(HTTPBasic())):
        """Basic Authentication"""
        correct_username = secrets.compare_digest(credentials.username, base64.b64decode(self.storage[0]).decode('ascii'))
        correct_password = secrets.compare_digest(credentials.password, base64.b64decode(self.storage[1]).decode('ascii'))
        if (correct_username and correct_password): return credentials.username
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect email or password",
                headers={"WWW-Authenticate": "Basic"},)
    ## For API
    def weak_authen_check(self, input_dict):
        """
        Sub Authentication for API in order to use with GCP Authentication
        , only use this function when 'api_weak_authen' is True
        """
        correct_username = input_dict.get('username','') == base64.b64decode(self.storage[0]).decode('ascii')
        correct_password = input_dict.get('password','') == base64.b64decode(self.storage[1]).decode('ascii')
        if (correct_username and correct_password): return 'ok'
        raise HTTPException(status_code=status.HTTP_401_UNAUTHORIZED,
                detail="Incorrect email or password",
                headers={"WWW-Authenticate": "Basic"},)

    def get_weak_authen_check(self, request: Request) :
        params = request.query_params
        self.weak_authen_check(params)

    def post_weak_authen_check(self, payload: dict):
        self.weak_authen_check(payload)

    def gen_doc(self) :           
        """Create Auth Doc"""
        @self.app.get("/openapi.json", include_in_schema=False)
        async def openapi(username: str = Depends(self.basic_authen_check)):
            return get_openapi(title=self.app.title, version=self.app.version, routes=self.app.routes)
        @self.app.route('/')
        @self.app.get("/docs", include_in_schema=False)
        async def get_swagger_documentation(username: str = Depends(self.basic_authen_check)):
            return get_swagger_ui_html(openapi_url="/openapi.json", title="docs")
        @self.app.get("/redoc", include_in_schema=False)
        async def get_redoc_documentation(username: str = Depends(self.basic_authen_check)):
            return get_redoc_html(openapi_url="/openapi.json", title="docs")

    def create_post(self, function, name, tags = [], example = {}, callback = 'default') :
        """Create POST Method for FastAPI"""
        # Assign Default Callback
        if (type(callback) == str) : 
            if callback == 'default' : callback = self.callback
        @self.app.post(f'/{name}' , tags = tags)
        async def post_function(  payload : dict = Body(... , example = example)
                                , username : str = Depends( self.post_weak_authen_check if self.api_weak_authen 
                                                            else self.basic_authen_check)
                                ) :
            try :
                return function(payload)
            except Exception as e :
                print(traceback.format_exc())
                if callback == None : 
                    return_error = {"error": str(traceback.format_exc())}
                else : 
                    return_error = callback({ 'traceback' : traceback.format_exc() 
                                            , 'exception' : e 
                                            , 'input' : payload})
                return JSONResponse(status_code=500
                                    , content=return_error)


    def create_get(self, function, name, tags = [], example = {}, callback = 'default') :
        """Create GET Method for FastAPI"""
        # Assign Default Callback
        if (type(callback) == str) : 
            if callback == 'default' : callback = self.callback
        # query_model = create_model("Query", **query_params)
        @self.app.get(f'/{name}' , tags = tags)
        async def get_function(  
                                # params: query_model = Depends()
                                request: Request
                                , username : str = Depends( self.get_weak_authen_check if self.api_weak_authen 
                                                            else self.basic_authen_check)
                                ) :
            try :
                params = request.query_params
                return function(params)
            except Exception as e :
                print(traceback.format_exc())
                if callback == None : 
                    return_error = {"error": str(traceback.format_exc())}
                else : 
                    return_error = callback({ 'traceback' : traceback.format_exc() 
                                            , 'exception' : e 
                                            , 'input' : request.query_params})
                return JSONResponse(status_code=500
                                    , content=return_error)
            
    def run(self, port = 8080, host = "0.0.0.0") :
        """To run FastAPI"""
        self.gen_doc()
        uvicorn.run(self.app, port=int(os.environ.get("PORT", port)), host = host)



import paramiko

class lazy_sftp_connect :
    """SFTP connection for lazy_SFTP"""
    def __init__(self, host_name, username, password, port) :
        self.host_name, self.username, self.password, self.port = host_name, username, password,port
        
    def __enter__(self) : 
        self.transport = paramiko.Transport((self.host_name,self.port))
        # Auth    
        self.transport.connect(None,self.username,self.password)
        # Go!    
        self.sftp = paramiko.SFTPClient.from_transport(self.transport)
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb) :
        if self.sftp: self.sftp.close()
        if self.transport: self.transport.close()

            
class lazy_SFTP :   
    """lazy_SFTP to interact with SFTP server"""
    def __init__(self, host_name, username, password, port = 22) :
        self.host_name, self.username, self.password, self.port = host_name, username, password,port
        self.con = lazy_sftp_connect(  host_name = self.host_name 
                                     , username = self.username
                                     , password = self.password
                                     , port = self.port)
        
    def listdir(self, path='.') :
        with self.con as connect :
            out = connect.sftp.listdir(path=path)
        return out
    
    def download(self, remotepath, localpath) :
        with self.con as connect :
            connect.sftp.get(localpath = localpath, remotepath = remotepath)

    def upload(self, remotepath, localpath) :
        with self.con as connect :
            connect.sftp.put(localpath = localpath, remotepath = remotepath)


import socket
from time import sleep

class lazy_TCP :
    def __init__(self, host, port, delay_close = 0) :
        self.host, self.port, self.delay_close = host, port, delay_close/1000
    
    def send(self, send_message, auto_encode = True, wait_return = False) :
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((self.host, self.port))
            if auto_encode : send_message = str.encode(send_message)
            s.sendall(send_message)
            if wait_return : data = s.recv(1024)
            else : data = None
            sleep(self.delay_close)
            return data
    
    def listen(self, debug = False, send_return = True) :
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s :
            s.bind((self.host, self.port))
            s.listen()
            conn, addr = s.accept()
            with conn:
                if debug : print(f'connect by {addr}')
                data = conn.recv(1024)
                if debug : print('receive :',data)
                if send_return : conn.sendall(data)
                sleep(self.delay_close)
        return data


from sklearn.preprocessing import LabelEncoder
from sklearn.tree import export_text
import pandas as pd
import numpy as np
import os
from joblib import dump, load

def create_encoder(df_in , col_in, folder_in , debug = False) :
    if not os.path.exists(folder_in) : 
        if debug : print('Folder {} created'.format(folder_in))
        os.mkdir(folder_in)
    for i_in in col_in :
        en_encoder = LabelEncoder()
        en_encoder.fit(df_in[i_in].astype('str').astype('category'))
        dump(en_encoder, '{}/en__{}__.pkl'.format(folder_in , i_in))
        del en_encoder

def encode_col(df_in , col_in , folder_in , debug = False) :
    df_out = df_in.copy()
    for i_in in col_in :
        en_encoder = load('{}/en__{}__.pkl'.format(folder_in, i_in))
        all_class = list(en_encoder.classes_)
        df_out[i_in] = df_out[i_in].astype('str').astype('category').apply(lambda j_in : 
                                                             en_encoder.transform([j_in])[0] if j_in in all_class 
                                                             else len(all_class) )
        if debug : print(i_in , ':' , len(all_class) ,'Classes')
    return df_out

def decode_col(df_in , col_in , folder_in , debug = False) :
    df_out = df_in.copy()
    for i_in in col_in :
        en_encoder = load('{}/en__{}__.pkl'.format(folder_in, i_in))
        all_class = list(en_encoder.classes_)
        df_out[i_in] = df_out[i_in].astype('int').apply(lambda j_in : 
                                                             en_encoder.inverse_transform([j_in])[0] if j_in < len(all_class) 
                                                             else 'unknown' )
        if debug : print(i_in , ':' , len(all_class) ,'Classes')
    return df_out

def create_lag(df_in , col_in , lag_range , lag_name = 'lag' , drop_null = True, debug = False) :
    df_out = df_in.copy()
    lag_col = []
    if lag_range >= 0 : loop_range = range(0,lag_range)
    elif lag_range < 0 : loop_range = range(lag_range - 1, -1)
    for i_in in loop_range :
        if debug : print(i_in + 1)
        df_buf = df_out[col_in].shift(i_in + 1)
        col_buf = ['{}_{}_{}'.format(j_in, lag_name , str(i_in + 1).replace('-','m')) for j_in in df_buf.columns]
        df_buf.columns = col_buf
        df_out = pd.concat([df_out,df_buf] , axis = 1)
        lag_col += col_buf
    if drop_null : df_out = df_out[df_out[lag_col].notnull().min(axis = 1) != 0]
    return df_out

class lazy_tree_logic :
    """Explan Decision Tree Logic, Need More Optimize and Naming"""
    def __init__(self, model, feature_list, max_depth = 1000, spacing = 1, decimals = 4) :
        list_logic = export_text(model, feature_names=feature_list, max_depth = max_depth, spacing=spacing,decimals=decimals)
        list_logic = list_logic.split('\n')
        df_logic = pd.DataFrame({'logic' : list_logic})
        df_logic['depth'] = df_logic['logic'].apply(lambda x : len([c for c in x if c in '|']))
        df_logic = df_logic[df_logic['depth'] > 0]
        df_logic['begin'] = (df_logic['depth'] <= df_logic['depth'].shift(1)) | (df_logic['depth'] == 1)
        df_logic['end'] = (df_logic['depth'] >= df_logic['depth'].shift(-1)) | (df_logic['depth'] == max(df_logic['depth']))
        df_logic['end'].iloc[-1] = True
        df_logic['group'] = df_logic['begin'].cumsum()
        df_logic['logic'] = df_logic['logic'].apply(lambda x : str(x).split('|-')[-1][1:])
        df_logic['logic'] = np.where(df_logic['end'], df_logic['logic'], df_logic['logic'].str.split(' '))
        df_logic['feature'] = np.where(df_logic['end'], np.NaN, df_logic['logic'].apply(lambda x : x[0]))
        df_logic['operator'] = np.where(df_logic['end'], np.NaN, df_logic['logic'].apply(lambda x : x[1]))
        df_logic['value'] = np.where(df_logic['end'], np.NaN, df_logic['logic'].apply(lambda x : x[-1]))
        df_logic['output'] = np.where(df_logic['end'], df_logic['logic'], np.NaN)
        all_logic = pd.DataFrame()
        for i in df_logic['group'].unique() :
            buffer = all_logic.append(df_logic[df_logic['group'] == i])
            buffer = buffer[buffer['depth'] <= buffer[buffer['group'] == i]['depth'].max()]
            grouplist = list(buffer['group'].unique())
            grouplist.reverse()
            for j in grouplist :
                if j in list(buffer['group']) :
                    buffer = buffer[(~buffer['depth'].isin(buffer[buffer['group'] == j]['depth']))|(buffer['group'] >= j)]
            buffer['group'] = int(i)
            all_logic = all_logic.append(buffer)
        self.logic_str = list_logic[:]
        self.feature_list = feature_list[:]
        self.logic_frame = all_logic.reset_index(drop = True)
    
    def explain(self, x, logic = None, asframe = False , asstring = False):
        if logic == None : logic = self.logic_frame
        operator_dict = {'>' : lambda x,y : x > y,'>=' : lambda x,y : x >= y
                        ,'<' : lambda x,y : x < y,'<=' : lambda x,y : x <= y
                        ,'==' : lambda x,y : x == y}
        cal = logic.merge(pd.DataFrame({'actual' : x}).reset_index()
                          , how = 'inner' , left_on = 'feature' , right_on = 'index')
        cal['check'] = cal.apply(lambda x : operator_dict[x['operator']](float(x['actual']),float(x['value'])), axis = 1)
        cal = cal.groupby(['group']).agg({"check": 'sum',"feature": 'count'})
        cal = cal[cal['check'] == cal['feature']]
        if len(cal) != 1 :
            raise Exception("Error Found matching {} Cases.\nSuggest to try increase max_depth or decimals".format(len(cal)))
        elif asframe :
            return logic[logic['group'] == cal.index[0]]
        elif asstring :
            return str(list(logic[logic['group'] == cal.index[0]]['logic']))
        else :
            return list(logic[logic['group'] == cal.index[0]]['logic'])


import pandas as pd
from sqlalchemy import create_engine , inspect, text
import os
from urllib.parse import quote_plus
from datetime import datetime


class lazy_SQL :
    """interact with SQL : Update 2022-06-03

    :sql_type = string - type of database, support MSSQL, MYSQL, POSTGRESQL, SQLITE, BIGQUERY

    :host_name = string 
       - for MSSQL, MYSQL, POSTGRESQL : database's hostname or ip address
       - for SQLITE : location of db file
       - for BIGQUERY : project id on GCP

    : database_name = string
       - for MSSQL, MYSQL, POSTGRESQL : database's name
       - for SQLITE : Leave as ''
       - for BIGQUERY : dataset's name

    : user = string - User for MSSQL, MYSQL, POSTGRESQL

    : password = string - password for MSSQL, MYSQL, POSTGRESQL

    : credentials_path = string - path to GCP credentials file (json)

    more sample of this class at https://github.com/chanon-kr/Shared_Function/blob/main/samples/database.ipynb
    """
    def __init__(self, sql_type, host_name, database_name, user = '', password = '' , credentials_path = None
                , chunksize = 150, partition_size = 5000, parallel_dump = False, max_parallel = 2, mute = False, **kwargs):
        """Create connection to SQL"""
        sql_type = sql_type.upper()
        self.sql_type = sql_type
        self.chunksize = int(chunksize)
        self.mute = mute
        password = quote_plus(str(password))
        if sql_type == 'BIGQUERY' : self.method = None
        else : self.method = 'multi'
        self.partition_size = int(partition_size)
        self.parallel_dump = parallel_dump
        if parallel_dump : print('Parallel dump has been remove since 0.3.18, this parameter will remove in 0.3.19')
        self.max_parallel = int(max_parallel)
        type_dic = {
                    'MSSQL' : ['mssql', 'pytds', '1433','[',']','EXEC'],
                    'MYSQL' : ['mysql', 'pymysql','3306','`','`','CALL'],
                    'POSTGRESQL' : ['postgresql', 'psycopg2','5432','"','"','SELECT * FROM '],
                    'SQLITE' : ['sqlite', 'sqlite', '','[',']','EXEC'],
                    'BIGQUERY' : ['bigquery', 'bigquery','','`','`','CALL'],
                    }
        
        if (sql_type == 'SQLITE') & (chunksize == 150) :
            self.chunksize = 50

        if type_dic.get(sql_type,'Error') == 'Error' :
            raise Exception("Please Insert Type of your Database with these range \n{}".format(list(type_dic.keys())))
        
        if (kwargs.get('driver',None) == 'pyodbc') & (kwargs.get('parameter', None) == None) :
            raise Exception("Please Insert ODBC's Driver in parameter agument\nfor example, for MSSQL : parameter = 'driver=SQL+Server'")
        
        if kwargs.get('parameter', None) == None : additional_param = ''
        else : additional_param = '?' + kwargs['parameter']

        self.begin_name, self.end_name, self.call_SP = type_dic[sql_type][3], type_dic[sql_type][4], type_dic[sql_type][5]
        
        if (sql_type == 'SQLITE') & (kwargs.get('driver',None) == None) :
            connection_str = """{}:///{}{}""".format(type_dic[sql_type][0] , host_name , additional_param)
        elif sql_type == 'BIGQUERY' :
            connection_str = "{}://{}/{}".format(type_dic[sql_type][0],host_name,database_name)
        else :
            connection_str = """{}+{}://{}:{}@{}:{}/{}{}""".format(type_dic[sql_type][0],kwargs.get('driver',type_dic[sql_type][1])
                                                            ,user,password,host_name,kwargs.get('port',type_dic[sql_type][2])
                                                            ,database_name , additional_param)
        
        if (credentials_path != '') & (sql_type == 'BIGQUERY') :
            from google.oauth2 import service_account
            self.project_id = host_name
            # self.dataset = self.begin_name + database_name + self.end_name + '.'
            self.dataset = database_name + '.'
            if credentials_path == None : 
                self.credentials, self.credentials_path = None, None
                self.engine = create_engine(connection_str)
            else : 
                self.engine = create_engine(connection_str, credentials_path = credentials_path)
                self.credentials = service_account.Credentials.from_service_account_file(credentials_path)
                self.credentials_path = credentials_path
            if not self.mute : print(pd.read_gbq("""SELECT 'Connection OK'""",project_id = self.project_id,credentials = self.credentials).iloc[0,0]) 
        else :
            self.credentials = ''
            self.credentials_path  = None
            self.dataset = ''
            self.engine = create_engine(connection_str)
            if not self.mute : 
                with self.engine.connect() as conn :
                    print(pd.read_sql_query("""SELECT 'Connection OK'""", con = conn).iloc[0,0]) 

    def read(self, table_name_in, condition_in = '', SP = False,raw = False, param = '', columns_list = []):
        """Read Table or View or Store Procedure"""
        columns_read = ''
        if len(columns_list) == 0 :columns_read = '*'
        else : 
            for i in columns_list :
                if columns_read != '' : columns_read += ', '
                columns_read += '{}{}{}.{}{}{}'.format(self.begin_name,table_name_in,self.end_name  ,  self.begin_name,i,self.end_name)
        if SP :
            if self.sql_type == 'POSTGRESQL' : return 'SP/Function None Avaliable for PostgreSQL, right now'
            sql_q = self.call_SP + ' ' + self.dataset + table_name_in + ' '
            if param == '' : raise Exception("Please insert SP's parameter")
            elif type(param) != dict : raise Exception("param must be dict")
            else :
                n = 0
                for i in param.keys() :
                    if type(param[i]) == str : param[i] = """'{}'""".format(param[i])
                    if self.sql_type in ['MYSQL','POSTGRESQL','BIGQUERY'] : 
                        if n == 0 : sql_q += ' (' 
                        else : sql_q += ' , '
                        sql_q += """ {}""".format(param[i])
                        if n + 1 == len(param.keys()) : sql_q += ') ' 
                    elif self.sql_type in ['MSSQL'] : 
                        if n > 0 : sql_q += ' , '
                        sql_q += """ {} = {}""".format(i, param[i])
                    else :
                        if n > 0 : sql_q += ' , '
                        sql_q += """ {} = {}""".format(i, param[i])
                    n += 1
        elif raw : sql_q = str(table_name_in)
        else :
            sql_q = """SELECT {} FROM {}{}{}{}""".format(columns_read, self.dataset,self.begin_name,table_name_in,self.end_name)
            if not condition_in == '' :
                sql_q += """ WHERE """ + condition_in
        if (self.credentials != '') & (self.sql_type == 'BIGQUERY') :
            return pd.read_gbq(sql_q,project_id = self.project_id,credentials = self.credentials)
        else :
            with self.engine.connect() as conn :
                out_ = pd.read_sql_query(sql_q , con = conn)
            return out_

    # def sub_dump(self, df_in,table_name_in,mode_in) :
    #     """normal to_sql for dask's delayed in dump_main"""
    #     if (self.credentials != '') & (self.sql_type == 'BIGQUERY') :
    #         # print('{}{}'.format(self.dataset , table_name_in)) # For Debug
    #         df_in.to_gbq('{}{}'.format(self.dataset , table_name_in),project_id = self.project_id
    #                                     ,credentials = self.credentials , if_exists = mode_in) #, reauth = True
    #     else :
    #         with self.engine.connect() as conn :
    #             df_in.to_sql(table_name_in, con = conn, index = False,if_exists = mode_in,chunksize = self.chunksize, method = self.method)
    #     return len(df_in)

    def dump_main(self, df_in, table_name_in ,mode_in, partition_size = '', con = '') :
        """Divide and Dump Dataframe Into Database"""
        con = self.engine if con == '' else con
        partition_size = self.partition_size if partition_size == '' else partition_size
        i, j, sum_len, df_length = 0, 1, 0, len(df_in)
        while i < df_length :
            i_next = i + partition_size
            # sum_len += self.sub_dump(df_in.iloc[i:i_next,:],table_name_in,mode_in) 
            df_dump = df_in.iloc[i:i_next,:]
            sum_len += len(df_dump)
            if (self.credentials != '') & (self.sql_type == 'BIGQUERY') :
                # print('{}{}'.format(self.dataset , table_name_in)) # For Debug
                df_dump.to_gbq('{}{}'.format(self.dataset , table_name_in),project_id = self.project_id
                                            ,credentials = self.credentials , if_exists = mode_in) #, reauth = True
            else :
                # with self.engine.connect() as con :
                df_dump.to_sql(table_name_in, con = con, index = False,if_exists = mode_in,chunksize = self.chunksize, method = self.method)
            j += 1
            i += partition_size
        return sum_len

    def write_logic_sql(self, key , value_in):
        """Write SQL Condition Query that include >, <, ="""
        #logic_query = self.begin_name+ str(key) + self.end_name + ' ' + value_in['logic'] + ' ' # For Debug
        value_in['type'] = value_in.get('type','')
        if value_in.get('value','there is no value') != 'there is no value' :
            if ('date' in value_in['type']) | ('time' in value_in['type']):  # Datetime will need '' in SQL Statement
                logic_query = 'CAST(' + self.begin_name+ str(key) + self.end_name + ' AS {}) '.format(value_in['type']) + value_in['logic'] + ' '
                logic_query += """'""" + str(value_in['value']) + """'"""
            else : 
                logic_query = self.begin_name+ str(key) + self.end_name + ' ' + value_in['logic'] + ' ' + str(value_in['value'])
        else : raise Exception("Please insert value data")  # will get error if not specific value
        return logic_query

    def dump_whole(self, df_in, table_name_in , fix_table = False, debug = False) :
        """Delete exists table and replace with new df"""
        result = {}
        with self.engine.connect() as con : 
            trans = con.begin()
            try :
                if inspect(self.engine).has_table(table_name_in) :
                    if fix_table : 
                        if not self.mute : print('Delete Existing data from Table at ',pd.Timestamp.now())
                        if self.credentials_path != None : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials_path
                        # with self.engine.connect() as con :
                        con.execute("""DELETE FROM {}{}{}""".format(self.begin_name,table_name_in,self.end_name))
                    else :
                        if not self.mute : print('Drop Existing Table at ',pd.Timestamp.now())
                        if self.credentials_path != None : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials_path
                        # with self.engine.connect() as con :
                        con.execute("""DROP TABLE {}{}{}""".format(self.begin_name,table_name_in,self.end_name))
                else :
                    if not self.mute : print('Do not have table to delete at',pd.Timestamp.now())
                #Dump df_in to database
                if not self.mute : print('Dump data to ',table_name_in,' Begin ',pd.Timestamp.now())
                self.dump_main( df_in, table_name_in ,mode_in = 'append', con = con) 
                trans.commit()
                if not self.mute : print('Dump data to ',table_name_in,' End ',pd.Timestamp.now())
                result['result'] = True
            except Exception as e :
                print(e)
                if not self.mute : print('Error During Dump to ',table_name_in,' Begin Rollback ',pd.Timestamp.now())
                trans.rollback()
                if not self.mute : print('Rollback ',table_name_in,' End ',pd.Timestamp.now())
                result['result'], result['error'] = False, e
            return result


    def write_in_sql(self,df_in , key) :
        """Write SQL Condition Query 'in (x,x,x)'"""
        filter_filter = tuple(df_in[key].fillna('Will BE rEpLaCe wItH NULL').astype('str').unique())
        if len(filter_filter) == 1 : 
            if 'Will BE rEpLaCe wItH NULL' in filter_filter :
                logic_query = '({}{}{} IS NULL'.format(self.begin_name , key, self.end_name) + ')'
            else :
                filter_filter = str(filter_filter).replace(',)',')')  # tuple with 1 value will be ( x , ) => need to convert
                if self.sql_type != 'BIGQUERY':
                    logic_query = self.begin_name + key  + self.end_name + ' in ' + filter_filter
                else :
                    logic_query = 'CAST(' + self.begin_name + key  + self.end_name + ' AS STRING) in ' + filter_filter
        elif len(filter_filter) > 1 :
            filter_filter = str(filter_filter) # tuple with > 1 values will be ( x, y, z) which can be use in SQL
            if self.sql_type != 'BIGQUERY':
                logic_query = self.begin_name + key  + self.end_name + ' in ' + filter_filter
            else :
                logic_query = 'CAST(' + self.begin_name + key  + self.end_name + ' AS STRING) in ' + filter_filter
        else : logic_query = '' # Return Nothing
        #print(logic_query)
        if "'Will BE rEpLaCe wItH NULL'" in logic_query :
            #print(logic_query) # For Debug
            logic_query = logic_query.replace(", 'Will BE rEpLaCe wItH NULL'",'')
            #print(logic_query) # For Debug
            logic_query = logic_query.replace("'Will BE rEpLaCe wItH NULL',",'')
            logic_query = logic_query.replace("'Will BE rEpLaCe wItH NULL'",'')
            logic_query = '(' + logic_query + ' OR {}{}{} IS NULL'.format(self.begin_name , key, self.end_name) + ')'
        return logic_query

    def delete_old_data(self, df_in, table_name_in, list_key, math_logic, debug, con = '') :
        """Sub Function in dump_replace to delete row in target table"""
        #Create SQL Query for Delete
        sql_q = """DELETE FROM {}{}{} where """.format(self.begin_name,table_name_in,self.end_name)

        if math_logic != '' : logic_list = list(math_logic.keys())
        else : logic_list = ['There_is_no_logic_math_need_to_write'] # column in logic_list will never be true

        #for single key
        if 'str' in str(type(list_key))  :  list_key = [list_key]
         
        if 'list' in str(type(list_key)) :
            if not min(x in df_in.columns for x in list_key) :
                raise Exception("Some keys are not in df's columns")
            i = 0
            while i < len(list_key) :
                filter_name = list_key[i]
                if filter_name in logic_list :
                    sql_q += self.write_logic_sql(filter_name, math_logic[filter_name])
                else :
                    sql_q += self.write_in_sql(df_in,filter_name)
                i += 1
                if i < len(list_key) : sql_q += ' and '
        #if key is not either string or list
        else :
            raise Exception("List of Key must be string or list")
        
        #Delete exiting row from table
        if inspect(self.engine).has_table(table_name_in) :
            if not self.mute : print('Start delete old data at',pd.Timestamp.now())
            if debug : print(sql_q)
            if self.credentials_path != None : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials_path
            # with self.engine.connect() as con :
            con.execute(sql_q)
            if not self.mute : print('Delete Last '+str(list_key)+' at',pd.Timestamp.now())
        else :
            if not self.mute : print('Do not have table to delete at',pd.Timestamp.now())


    def create_merge_statement(self, sql_type, DATASET , DESTINATION_TABLE ,TEMP_TABLE ,SCHEMA, list_key, list_column, list_value) :
        key_columns_string = ' , '.join([i for i in list_key])
        now = datetime.now().strftime('%Y%m%d%H%M%S')
        create_index, drop_index = [], []
        for i in [DESTINATION_TABLE , TEMP_TABLE] :
            create_index.append(f"CREATE UNIQUE INDEX topping{i}{now} ON {i} ({key_columns_string})")
            if sql_type == 'postgresql' : drop_index.append(f"DROP INDEX topping{i}{now};")
            else : drop_index.append(f"DROP INDEX topping{i}{now} ON {i} ;")

        if len(list_value) == 0 : update_value_str = " , ".join([f"""{i} = SOURCE.{i}""" for i in list_key])
        else : update_value_str = " , ".join([f"""{i} = SOURCE.{i}""" for i in list_value])

        if sql_type == 'MYSQL' : 
            columns_list_str = " , ".join([i for i in list_column])
            columns_list_str2 = " , ".join([f'SOURCE.{i}' for i in list_column])
            sql_statement = f' INSERT INTO {DESTINATION_TABLE} ({columns_list_str})'
            sql_statement += f' SELECT {columns_list_str2} FROM {TEMP_TABLE}  AS SOURCE'
            sql_statement += f' ON DUPLICATE KEY UPDATE {update_value_str}'
            return sql_statement, create_index, drop_index

        if sql_type == 'MSSQL' : DATASET = f'{DATASET}.{SCHEMA}'
        
        if sql_type == 'POSTGRESQL' : sql_statement = f"""MERGE INTO {DESTINATION_TABLE} TARGET USING {TEMP_TABLE} SOURCE ON """
        else : sql_statement = f"""MERGE {DATASET}.{DESTINATION_TABLE} TARGET USING {DATASET}.{TEMP_TABLE} SOURCE ON """

        sql_statement += " AND ".join([f"""TARGET.{i} = SOURCE.{i}""" for i in list_key])
        sql_statement += f""" WHEN MATCHED THEN UPDATE SET """
        sql_statement += update_value_str

        if sql_type == 'POSTGRESQL' : sql_statement += f""" WHEN NOT MATCHED THEN INSERT ("""
        else : sql_statement += f""" WHEN NOT MATCHED BY TARGET THEN INSERT ("""

        sql_statement += " , ".join([i for i in list_column])
        sql_statement += f""") VALUES ("""      
        sql_statement += " , ".join([i for i in list_column])
        sql_statement += f""") ;"""

        return sql_statement, create_index, drop_index

    def upsert(self, df_in, table_name_in, list_key, debug = False, indexed = False, keep_temp = False) :
        list_column = df_in.columns
        list_value = [i for i in list_column if i not in list_key]
        result = {}
        if self.sql_type == 'MYSQL' : indexed = True
        TEMP_TABLE = f"{table_name_in}topping{datetime.now().strftime('%Y%m%d%H%M%S')}" 
        sql_statement, create_index, drop_index = self.create_merge_statement(
                                      sql_type = self.sql_type, DATASET = self.database_name
                                    , DESTINATION_TABLE = table_name_in
                                    , TEMP_TABLE = TEMP_TABLE
                                    , SCHEMA = 'dbo', list_key = list_key, list_column = list_column, list_value = list_value)
        with self.engine.connect() as con :
            try : 
                trans = con.begin()
                self.dump_whole( df_in , TEMP_TABLE)
                if indexed : 
                    for i in create_index : 
                        if debug : print(i)
                        con.execute(text(i))
                if debug : print(sql_statement)
                con.execute(text(sql_statement))
                if indexed : 
                    for i in drop_index : 
                        if debug : print(i)
                        con.execute(text(i))
                if not keep_temp : con.execute(text(f'DROP TABLE {TEMP_TABLE}'))
                else : result['temp'] = TEMP_TABLE
                trans.commit()
                result['result'] = True
            except Exception as e :
                if debug : print(e)
                trans.rollback()
                result['result'], result['error'] = False, e
        return result

    def dump_replace(self, df_in, table_name_in, list_key, math_logic = '', partition_delete = 50000, debug = False):
        """Delete exists row of table in database with same key(s) as df and dump df append to table"""
        result = {}
        with self.engine.connect() as con : 
            trans = con.begin()
            try :
                if len(df_in) == 0 : 
                    if not self.mute : print('No Data to dump into',table_name_in,' End ',pd.Timestamp.now())
                else :
                    if len(df_in) <= partition_delete : 
                        self.delete_old_data(df_in, table_name_in, list_key, math_logic, debug, con = con)
                    else :
                        i, df_length = 0, len(df_in)
                        while i < df_length :
                            i_next = i + partition_delete
                            if not self.mute : print("Processing {}-{} row from {} rows".format(i + 1, i_next, df_length))
                            self.delete_old_data(df_in.iloc[i:i_next,:],table_name_in, list_key, math_logic, debug, con = con)
                            i += partition_delete
                    #Dump df_in append to database
                    self.dump_main(df_in, table_name_in ,mode_in = 'append', con = con)
                    trans.commit()
                    if not self.mute : print('Dump data to ',table_name_in,' End ',pd.Timestamp.now())
                    result['result'] = True
            except Exception as e :
                print(e)
                if not self.mute : print('Error During Dump to ',table_name_in,' Begin Rollback ',pd.Timestamp.now())
                trans.rollback()
                if not self.mute : print('Rollback ',table_name_in,' End ',pd.Timestamp.now())
                result['result'], result['error'] = False, e
            return result

    def dump_new(self, df_in, table_name_in, list_key , debug = False) :
        """Delete exists row of df that has same key(s) as table and dump df_in append to table"""
        result = {}
        with self.engine.connect() as con : 
            trans = con.begin()
            try :
                # Check if Table existing or not
                if inspect(self.engine).has_table(table_name_in) :
                    if not self.mute : print('Start Filter Existing data from df at ',pd.Timestamp.now())
                    df_out = df_in.copy()
                    #for single key
                    if 'str' in str(type(list_key))  :  list_key = [list_key]
                    if 'list' in str(type(list_key)) :
                        if not min(x in df_out.columns for x in list_key) :
                            raise Exception("Some keys are not in df's columns")
                        i = 0
                        df_out['key_sql_filter'] = ''
                        sql_q = 'SELECT DISTINCT '
                        while i < len(list_key) :
                            filter_name = list_key[i]
                            if i != 0 : sql_q += ' , '
                            df_out['key_sql_filter'] = df_out['key_sql_filter'].astype('str') + df_out[filter_name].astype('str')
                            sql_q += self.begin_name +  filter_name + self.end_name + ' ' 
                            i += 1
                        sql_q += ' FROM ' + self.dataset + table_name_in
                        if debug : print(sql_q)
                        if (self.credentials != '') & (self.sql_type == 'BIGQUERY') :
                            filter_filter = pd.read_gbq(sql_q,project_id = self.project_id,credentials = self.credentials)
                        else :
                            # with self.engine.connect() as con :
                            filter_filter = pd.read_sql_query(sql_q, con = con)
                        # filter_filter = pd.read_sql_query(sql_q, con = self.engine)
                        filter_filter['key_sql_filter'] = ''
                        for i in filter_filter.columns :
                            if  i == 'key_sql_filter' : pass
                            else : filter_filter['key_sql_filter'] = filter_filter['key_sql_filter'].astype('str') + filter_filter[i].astype('str')
                        if debug : print(filter_filter.head())
                        filter_filter = filter_filter['key_sql_filter'].unique()
                        if debug : print('======filter======\n',filter_filter,'\n======in======\n',df_out.head())
                        logic_filter = ~df_out['key_sql_filter'].isin(filter_filter)
                        df_out = df_out.drop('key_sql_filter', axis = 1)
                    #if key is not either string or list
                    else :
                        raise Exception("List of Key must be string or list")
                    df_out = df_out[logic_filter]
                else : 
                    if not self.mute : print('Table not existing at ',pd.Timestamp.now())
                #Dump df_in append to database
                if not self.mute : print('Dump data to ',table_name_in,' Begin ',pd.Timestamp.now())
                self.dump_main( df_out, table_name_in ,mode_in = 'append', con = con) 
                trans.commit()
                if not self.mute : print('Dump data to ',table_name_in,' End ', pd.Timestamp.now())
                result['result'] = True
            except Exception as e :
                print(e)
                if not self.mute : print('Error During Dump to ',table_name_in,' Begin Rollback ',pd.Timestamp.now())
                trans.rollback()
                if not self.mute : print('Rollback ',table_name_in,' End ',pd.Timestamp.now())
                result['result'], result['error'] = False, e
            return result

class da_tran_SQL :
    """interact with SQL : Update 2021-11-25

    :sql_type = string - type of database, support MSSQL, MYSQL, POSTGRESQL, SQLITE, BIGQUERY

    :host_name = string 
       - for MSSQL, MYSQL, POSTGRESQL : database's hostname or ip address
       - for SQLITE : location of db file
       - for BIGQUERY : project id on GCP

    : database_name = string
       - for MSSQL, MYSQL, POSTGRESQL : database's name
       - for SQLITE : Leave as ''
       - for BIGQUERY : dataset's name

    : user = string - User for MSSQL, MYSQL, POSTGRESQL

    : password = string - password for MSSQL, MYSQL, POSTGRESQL

    : credentials_path = string - path to GCP credentials file (json)

    more sample of this class at https://github.com/chanon-kr/Shared_Function/blob/main/samples/database.ipynb
    """
    def __init__(self, sql_type, host_name, database_name, user = '', password = '' , credentials_path = None
                , chunksize = 150, partition_size = 5000, parallel_dump = False, max_parallel = 2, **kwargs):
        """Create connection to SQL"""
        sql_type = sql_type.upper()
        self.sql_type = sql_type
        self.chunksize = int(chunksize)
        password = quote_plus(str(password))
        if sql_type == 'BIGQUERY' : self.method = None
        else : self.method = 'multi'
        self.partition_size = int(partition_size)
        self.parallel_dump = parallel_dump
        if parallel_dump : print('Set to use parallel dump, there is a risk to use this method, use it with caution')
        self.max_parallel = int(max_parallel)
        type_dic = {
                    'MSSQL' : ['mssql', 'pymssql', '1433','[',']','EXEC'],
                    'MYSQL' : ['mysql', 'pymysql','3306','`','`','CALL'],
                    'POSTGRESQL' : ['postgresql', 'psycopg2','5432','"','"','SELECT * FROM '],
                    'SQLITE' : ['sqlite', 'sqlite', '','[',']','EXEC'],
                    'BIGQUERY' : ['bigquery', 'bigquery','','`','`','CALL'],
                    }
        
        if (sql_type == 'SQLITE') & (chunksize == 150) :
            self.chunksize = 50

        if type_dic.get(sql_type,'Error') == 'Error' :
            raise Exception("Please Insert Type of your Database with these range \n{}".format(list(type_dic.keys())))
        
        if (kwargs.get('driver',None) == 'pyodbc') & (kwargs.get('parameter', None) == None) :
            raise Exception("Please Insert ODBC's Driver in parameter agument\nfor example, for MSSQL : parameter = 'driver=SQL+Server'")
        
        if kwargs.get('parameter', None) == None : additional_param = ''
        else : additional_param = '?' + kwargs['parameter']

        self.begin_name, self.end_name, self.call_SP = type_dic[sql_type][3], type_dic[sql_type][4], type_dic[sql_type][5]
        
        if (sql_type == 'SQLITE') & (kwargs.get('driver',None) == None) :
            connection_str = """{}:///{}{}""".format(type_dic[sql_type][0] , host_name , additional_param)
        elif sql_type == 'BIGQUERY' :
            connection_str = "{}://{}/{}".format(type_dic[sql_type][0],host_name,database_name)
        else :
            connection_str = """{}+{}://{}:{}@{}:{}/{}{}""".format(type_dic[sql_type][0],kwargs.get('driver',type_dic[sql_type][1])
                                                            ,user,password,host_name,kwargs.get('port',type_dic[sql_type][2])
                                                            ,database_name , additional_param)
        
        if (credentials_path != '') & (sql_type == 'BIGQUERY') :
            from google.oauth2 import service_account
            self.project_id = host_name
            # self.dataset = self.begin_name + database_name + self.end_name + '.'
            self.dataset = database_name + '.'
            if credentials_path == None : 
                self.credentials, self.credentials_path = None, None
                self.engine = create_engine(connection_str)
            else : 
                self.engine = create_engine(connection_str, credentials_path = credentials_path)
                self.credentials = service_account.Credentials.from_service_account_file(credentials_path)
                self.credentials_path = credentials_path
            print(pd.read_gbq("""SELECT 'Connection OK'""",project_id = self.project_id,credentials = self.credentials).iloc[0,0]) 
        else :
            self.credentials = ''
            self.credentials_path  = None
            self.dataset = ''
            self.engine = create_engine(connection_str)
            print(pd.read_sql_query("""SELECT 'Connection OK'""", con = self.engine).iloc[0,0]) 

    def sub_dump(self, df_in,table_name_in,mode_in) :
        """normal to_sql for dask's delayed in dump_main"""
        if (self.credentials != '') & (self.sql_type == 'BIGQUERY') :
            # print('{}{}'.format(self.dataset , table_name_in)) # For Debug
            df_in.to_gbq('{}{}'.format(self.dataset , table_name_in),project_id = self.project_id
                                        ,credentials = self.credentials , if_exists = mode_in) #, reauth = True
        else :
            df_in.to_sql(table_name_in, con = self.engine, index = False,if_exists = mode_in,chunksize = self.chunksize, method = self.method)
        return len(df_in)

    def dump_main(self, df_in, table_name_in ,mode_in) :
        """Divide and Dump Dataframe Into Database"""
        if len(df_in) <= self.partition_size : sum_len = self.sub_dump(df_in,table_name_in,mode_in) 
        else :
            if self.parallel_dump : 
                print('About to use parallel dump, there is a risk to use this method, use it with caution')
                from dask import delayed
                dask_dump = delayed(self.sub_dump)
            i, j, sum_len, df_length = 0, 1, 0, len(df_in)
            while i < df_length :
                i_next = i + self.partition_size
                if self.parallel_dump :
                    sum_len += dask_dump(df_in.iloc[i:i_next,:],table_name_in,mode_in) 
                    if (j == self.max_parallel) | (i_next >= df_length) : 
                        sum_len, j = sum_len.compute(), 1
                        if sum_len < i_next - 1  : 
                            raise Exception("Parallel Dump Error Around row {} to row {}".format(int(i_next - self.partition_size*j),i_next))
                else : sum_len += self.sub_dump(df_in.iloc[i:i_next,:],table_name_in,mode_in) 
                j += 1
                i += self.partition_size
        return sum_len


    def read(self, table_name_in, condition_in = '', SP = False,raw = False, param = '', columns_list = []):
        """Read Table or View or Store Procedure"""
        columns_read = ''
        if len(columns_list) == 0 :columns_read = '*'
        else : 
            for i in columns_list :
                if columns_read != '' : columns_read += ', '
                columns_read += '{}{}{}.{}{}{}'.format(self.begin_name,table_name_in,self.end_name  ,  self.begin_name,i,self.end_name)
        if SP :
            if self.sql_type == 'POSTGRESQL' : return 'SP/Function None Avaliable for PostgreSQL, right now'
            sql_q = self.call_SP + ' ' + self.dataset + table_name_in + ' '
            if param == '' : raise Exception("Please insert SP's parameter")
            elif type(param) != dict : raise Exception("param must be dict")
            else :
                n = 0
                for i in param.keys() :
                    if type(param[i]) == str : param[i] = """'{}'""".format(param[i])
                    if self.sql_type in ['MYSQL','POSTGRESQL','BIGQUERY'] : 
                        if n == 0 : sql_q += ' (' 
                        else : sql_q += ' , '
                        sql_q += """ {}""".format(param[i])
                        if n + 1 == len(param.keys()) : sql_q += ') ' 
                    elif self.sql_type in ['MSSQL'] : 
                        if n > 0 : sql_q += ' , '
                        sql_q += """ {} = {}""".format(i, param[i])
                    else :
                        if n > 0 : sql_q += ' , '
                        sql_q += """ {} = {}""".format(i, param[i])
                    n += 1
        elif raw : sql_q = str(table_name_in)
        else :
            sql_q = """SELECT {} FROM {}{}{}{}""".format(columns_read, self.dataset,self.begin_name,table_name_in,self.end_name)
            if not condition_in == '' :
                sql_q += """ WHERE """ + condition_in
        if (self.credentials != '') & (self.sql_type == 'BIGQUERY') :
            return pd.read_gbq(sql_q,project_id = self.project_id,credentials = self.credentials)
        else :
            return pd.read_sql_query(sql_q , con = self.engine)

    def write_logic_sql(self, key , value_in):
        """Write SQL Condition Query that include >, <, ="""
        #logic_query = self.begin_name+ str(key) + self.end_name + ' ' + value_in['logic'] + ' ' # For Debug
        value_in['type'] = value_in.get('type','')
        if value_in.get('value','there is no value') != 'there is no value' :
            if ('date' in value_in['type']) | ('time' in value_in['type']):  # Datetime will need '' in SQL Statement
                logic_query = 'CAST(' + self.begin_name+ str(key) + self.end_name + ' AS {}) '.format(value_in['type']) + value_in['logic'] + ' '
                logic_query += """'""" + str(value_in['value']) + """'"""
            else : 
                logic_query = self.begin_name+ str(key) + self.end_name + ' ' + value_in['logic'] + ' ' + str(value_in['value'])
        else : raise Exception("Please insert value data")  # will get error if not specific value
        return logic_query

    def dump_whole(self, df_in, table_name_in , fix_table = False, debug = False) :
        """Delete exists table and replace with new df"""
        if table_name_in in list(self.engine.table_names()) :
            if fix_table : 
                print('Delete Existing data from Table at ',pd.Timestamp.now())
                if self.credentials_path != None : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials_path
                self.engine.execute("""DELETE FROM {}{}{}""".format(self.begin_name,table_name_in,self.end_name))
            else :
                print('Drop Existing Table at ',pd.Timestamp.now())
                if self.credentials_path != None : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials_path
                self.engine.execute("""DROP TABLE {}{}{}""".format(self.begin_name,table_name_in,self.end_name))
        else :
            print('Do not have table to delete at',pd.Timestamp.now())
        #Dump df_in to database
        print('Dump data to ',table_name_in,' Begin ',pd.Timestamp.now())
        self.dump_main( df_in, table_name_in ,mode_in = 'append') 
        print('Dump data to ',table_name_in,' End ',pd.Timestamp.now())

    def write_in_sql(self,df_in , key) :
        """Write SQL Condition Query 'in (x,x,x)'"""
        filter_filter = tuple(df_in[key].fillna('Will BE rEpLaCe wItH NULL').astype('str').unique())
        if len(filter_filter) == 1 : 
            if 'Will BE rEpLaCe wItH NULL' in filter_filter :
                logic_query = '({}{}{} IS NULL'.format(self.begin_name , key, self.end_name) + ')'
            else :
                filter_filter = str(filter_filter).replace(',)',')')  # tuple with 1 value will be ( x , ) => need to convert
                if self.sql_type != 'BIGQUERY':
                    logic_query = self.begin_name + key  + self.end_name + ' in ' + filter_filter
                else :
                    logic_query = 'CAST(' + self.begin_name + key  + self.end_name + ' AS STRING) in ' + filter_filter
        elif len(filter_filter) > 1 :
            filter_filter = str(filter_filter) # tuple with > 1 values will be ( x, y, z) which can be use in SQL
            if self.sql_type != 'BIGQUERY':
                logic_query = self.begin_name + key  + self.end_name + ' in ' + filter_filter
            else :
                logic_query = 'CAST(' + self.begin_name + key  + self.end_name + ' AS STRING) in ' + filter_filter
        else : logic_query = '' # Return Nothing
        #print(logic_query)
        if "'Will BE rEpLaCe wItH NULL'" in logic_query :
            #print(logic_query) # For Debug
            logic_query = logic_query.replace(", 'Will BE rEpLaCe wItH NULL'",'')
            #print(logic_query) # For Debug
            logic_query = logic_query.replace("'Will BE rEpLaCe wItH NULL',",'')
            logic_query = logic_query.replace("'Will BE rEpLaCe wItH NULL'",'')
            logic_query = '(' + logic_query + ' OR {}{}{} IS NULL'.format(self.begin_name , key, self.end_name) + ')'
        return logic_query

    def delete_old_data(self, df_in, table_name_in, list_key, math_logic, debug) :
        """Sub Function in dump_replace to delete row in target table"""
        #Create SQL Query for Delete
        sql_q = """DELETE FROM {}{}{} where """.format(self.begin_name,table_name_in,self.end_name)

        if math_logic != '' : logic_list = list(math_logic.keys())
        else : logic_list = ['There_is_no_logic_math_need_to_write'] # column in logic_list will never be true

        #for single key
        if 'str' in str(type(list_key))  :  list_key = [list_key]
         
        if 'list' in str(type(list_key)) :
            if not min(x in df_in.columns for x in list_key) :
                raise Exception("Some keys are not in df's columns")
            i = 0
            while i < len(list_key) :
                filter_name = list_key[i]
                if filter_name in logic_list :
                    sql_q += self.write_logic_sql(filter_name, math_logic[filter_name])
                else :
                    sql_q += self.write_in_sql(df_in,filter_name)
                i += 1
                if i < len(list_key) : sql_q += ' and '
        #if key is not either string or list
        else :
            raise Exception("List of Key must be string or list")
        
        #Delete exiting row from table
        if table_name_in in list(self.engine.table_names()) :
            print('Start delete old data at',pd.Timestamp.now())
            if debug : print(sql_q)
            if self.credentials_path != None : os.environ["GOOGLE_APPLICATION_CREDENTIALS"]= self.credentials_path
            self.engine.execute(sql_q)
            print('Delete Last '+str(list_key)+' at',pd.Timestamp.now())
        else :
            print('Do not have table to delete at',pd.Timestamp.now())

    def dump_replace(self, df_in, table_name_in, list_key, math_logic = '', partition_delete = 50000, debug = False):
        """Delete exists row of table in database with same key(s) as df and dump df append to table"""
        if len(df_in) == 0 : 
            print('No Data to dump into',table_name_in,' End ',pd.Timestamp.now())
        else :
            if len(df_in) <= partition_delete : 
                self.delete_old_data(df_in, table_name_in, list_key, math_logic, debug)
            else :
                i, df_length = 0, len(df_in)
                while i < df_length :
                    i_next = i + partition_delete
                    print("Processing {}-{} row from {} rows".format(i + 1, i_next, df_length))
                    self.delete_old_data(df_in.iloc[i:i_next,:],table_name_in, list_key, math_logic, debug)
                    i += partition_delete
            #Dump df_in append to database
            self.dump_main(df_in, table_name_in ,mode_in = 'append')
            print('Dump data to ',table_name_in,' End ',pd.Timestamp.now())

    def dump_new(self, df_in, table_name_in, list_key , debug = False) :
        """Delete exists row of df that has same key(s) as table and dump df_in append to table"""

        # Check if Table existing or not
        if table_name_in in list(self.engine.table_names()) :
            print('Start Filter Existing data from df at ',pd.Timestamp.now())
            
            df_out = df_in.copy()
            #for single key
            if 'str' in str(type(list_key))  :  list_key = [list_key]

            if 'list' in str(type(list_key)) :
                if not min(x in df_out.columns for x in list_key) :
                    raise Exception("Some keys are not in df's columns")
                i = 0
                df_out['key_sql_filter'] = ''
                sql_q = 'SELECT DISTINCT '
                while i < len(list_key) :
                    filter_name = list_key[i]
                    if i != 0 : sql_q += ' , '
                    df_out['key_sql_filter'] = df_out['key_sql_filter'].astype('str') + df_out[filter_name].astype('str')
                    sql_q += self.begin_name +  filter_name + self.end_name + ' ' 
                    i += 1
                sql_q += ' FROM ' + self.dataset + table_name_in
                if debug : print(sql_q)
                if (self.credentials != '') & (self.sql_type == 'BIGQUERY') :
                    filter_filter = pd.read_gbq(sql_q,project_id = self.project_id,credentials = self.credentials)
                else :
                    filter_filter = pd.read_sql_query(sql_q, con = self.engine)
                # filter_filter = pd.read_sql_query(sql_q, con = self.engine)
                filter_filter['key_sql_filter'] = ''
                for i in filter_filter.columns :
                    if  i == 'key_sql_filter' : pass
                    else : filter_filter['key_sql_filter'] = filter_filter['key_sql_filter'].astype('str') + filter_filter[i].astype('str')
                if debug : print(filter_filter.head())
                filter_filter = filter_filter['key_sql_filter'].unique()
                if debug : print('======filter======\n',filter_filter,'\n======in======\n',df_out.head())
                logic_filter = ~df_out['key_sql_filter'].isin(filter_filter)
                df_out = df_out.drop('key_sql_filter', axis = 1)

            #if key is not either string or list
            else :
                raise Exception("List of Key must be string or list")

            df_out = df_out[logic_filter]
            
        else : print('Table not existing at ',pd.Timestamp.now())

        #Dump df_in append to database
        print('Dump data to ',table_name_in,' Begin ',pd.Timestamp.now())
        self.dump_main( df_out, table_name_in ,mode_in = 'append') 
        print('Dump data to ',table_name_in,' End ', pd.Timestamp.now())